<?php
/**
* Copyright (C) 2017 Petr Hucik <petr@getdatakick.com>
*
* NOTICE OF LICENSE
*
* This source file is subject to the Academic Free License (AFL 3.0)
* that is bundled with this package in the file LICENSE.txt.
* It is also available through the world-wide-web at this URL:
* http://opensource.org/licenses/afl-3.0.php
* If you did not receive a copy of the license and are unable to
* obtain it through the world-wide-web, please send an email
* to license@getdatakick.com so we can send you a copy immediately.
*
* @author    Petr Hucik <petr@getdatakick.com>
* @copyright 2018 Petr Hucik
* @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
*/

namespace Revws;
use \Customer;
use \Db;
use \Shop;
use \RevwsReview;

class Visitor {
  const GUEST = 'guest';
  const CUSTOMER = 'customer';

  private $settings;
  private $type;
  private $id;
  private $firstName='';
  private $lastName='';
  private $email='';
  private $reactions = null;
  private $reviewedProducts = null;

  public function __construct($context, Settings $settings) {
    $this->settings = $settings;
    if ($context->customer->isLogged()) {
      $this->type = self::CUSTOMER;
      $this->id = (int)$context->customer->id;
      $customer = new Customer($this->id);
      $this->email = $customer->email;
      $this->firstName = $customer->firstname;
      $this->lastName = $customer->lastname;
    } else {
      $this->type = self::GUEST;
      $this->id = (int)$context->cookie->id_guest;
    }
  }

  public function isGuest() {
    return $this->type == self::GUEST;
  }

  public function isCustomer() {
    return $this->type == self::CUSTOMER;
  }

  public function getId() {
    return $this->id;
  }

  public function getType() {
    return $this->type;
  }

  public function getFirstName() {
    return $this->firstName;
  }

  public function getLastName() {
    return $this->lastName;
  }

  public function getEmail() {
    return trim($this->email);
  }

  public function hasReacted($reviewId, $reactionType) {
    $this->loadReactions();
    return isset($this->reactions[(int)$reviewId][$reactionType]);
  }

  public function getCustomerId() {
    return $this->isCustomer() ? $this->id : 0;
  }

  public function getGuestId() {
    return $this->isGuest() ? $this->id : 0;
  }

  public function hasWrittenReview($productId) {
    $this->loadReviews();
    return isset($this->reviewedProducts[$productId]);
  }

  public function getReviewedProducts() {
    $this->loadReviews();
    return array_keys($this->reviewedProducts);
  }

  private function loadReactions() {
    if (is_null($this->reactions)) {
      $conn = Db::getInstance(_PS_USE_SQL_SLAVE_);
      $table = _DB_PREFIX_ . 'revws_review_reaction';
      $this->reactions = [];
      $query = "SELECT * FROM $table WHERE id_customer = {$this->getCustomerId()} AND id_guest = {$this->getGuestId()}";
      foreach ($conn->executeS($query) as $row) {
        $review = (int)$row['id_review'];
        $type = $row['reaction_type'];
        $this->reactions[$review][$type] = true;
      }
    }
  }

  private function loadReviews() {
    if (is_null($this->reviewedProducts)) {
      $this->reviewedProducts = [];
      $visitorType = $this->getType();
      $visitorId = $this->getId();
      $reviews = RevwsReview::findReviews([
        'deleted' => false,
        $visitorType => (int)$visitorId
      ]);
      foreach($reviews['reviews'] as $rev) {
        $this->reviewedProducts[(int)$rev->id_product] = true;
      }
    }
  }

  public function getProductsToReview() {
    if ($this->isGuest()) {
      return [];
    }
    $customer = (int)$this->getCustomerId();
    $shop = (int)Shop::getContextShopID();
    $conn = Db::getInstance(_PS_USE_SQL_SLAVE_);
    $sql = ("
      SELECT d.product_id, o.date_add
        FROM "._DB_PREFIX_."orders o
        INNER JOIN "._DB_PREFIX_."order_detail d ON (o.id_order = d.id_order AND o.id_shop=d.id_shop)
        INNER JOIN "._DB_PREFIX_."product_shop p ON (p.id_product = d.product_id and p.id_shop = d.id_shop)
        LEFT JOIN  "._DB_PREFIX_."revws_review r ON (r.id_product = p.id_product)
        WHERE o.id_customer = $customer
          AND o.id_shop = $shop
          AND r.id_review IS NULL
        ORDER BY o.date_add DESC
    ");
    $data = [];
    foreach ($conn->executeS($sql) as $row) {
      $productId = (int)$row['product_id'];
      if (! isset($data[$productId])) {
        $data[$productId] = true;
      }
    }
    return array_keys($data);
  }

  public function hasPurchasedProduct($productId) {
    if ($this->isGuest()) {
      return false;
    }
    $customer = (int)$this->getCustomerId();
    $shop = (int)Shop::getContextShopID();
    $productId = (int)$productId;
    $conn = Db::getInstance(_PS_USE_SQL_SLAVE_);
    $sql = ("
      SELECT 1
        FROM "._DB_PREFIX_."orders o
        INNER JOIN "._DB_PREFIX_."order_detail d ON (o.id_order = d.id_order AND o.id_shop=d.id_shop)
        INNER JOIN "._DB_PREFIX_."product_shop p ON (p.id_product = d.product_id and p.id_shop = d.id_shop)
        WHERE o.id_customer = $customer
          AND o.id_shop = $shop
          AND o.delivery_date IS NOT NULL
          AND p.id_product = $productId
    ");
    $res = $conn->executeS($sql);
    return !empty($res);
  }

}
